"""
Module to simplify the use of Ffmpeg and to make
awesome things with simple methods.

Nice help: https://www.bannerbear.com/blog/how-to-use-ffmpeg-in-python-with-examples/
Official doc: https://www.ffmpeg.org/ffmpeg-resampler.html
More help: https://kkroening.github.io/ffmpeg-python/
Nice guide: https://img.ly/blog/ultimate-guide-to-ffmpeg/
Available flags: https://gist.github.com/tayvano/6e2d456a9897f55025e25035478a3a50

Interesting usage: https://stackoverflow.com/a/20325676
Maybe avoid writting on disk?: https://github.com/kkroening/ffmpeg-python/issues/500#issuecomment-792281072
"""
from yta_video_ffmpeg.handler import FfmpegHandler


__all__ = [
    'FfmpegHandler'
]


if __name__ == '__main__':
    from yta_video_ffmpeg.reader import FfmpegReader

    info = FfmpegReader.get_video_metadata('test_files/test_1.mp4')
    print(info.has_alpha)
    print(info.duration)


from yta_date import Date
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from yta_constants.video import FfmpegAudioCodec, FfmpegFilter, FfmpegPixelFormat, FfmpegVideoCodec, FfmpegVideoFormat
from typing import Union


class FfmpegFlag:
    """
    Class to simplify the way we push flags into
    the ffmpeg command.
    """

    overwrite: str = '-y'
    """
    Overwrite the output file if existing.

    Code: `-y`
    """

    @staticmethod
    def force_format(
        format: FfmpegVideoFormat
    ) -> str:
        """
        Force the output format to be the provided 'format'.

        Code: `-f {format}`
        """
        format = FfmpegVideoFormat.to_enum(format).value

        return f'-f {format}'
    
    @staticmethod
    def safe_routes(
        value: int
    ) -> str:
        """
        To enable or disable unsafe paths.

        Code: `-safe {value}`
        """
        ParameterValidator.validate_mandatory_int('value', value)
        # TODO: Check that 'value' is a number between -1 and 1

        return f'-safe {str(value)}'
    
    @staticmethod
    def input(
        input: str
    ) -> str:
        """
        To set the input (or inputs) we want.

        Code: `-i {input}`
        """
        ParameterValidator.validate_mandatory_string('input', input, do_accept_empty = False)

        return f'-i {input}'
    
    @staticmethod
    def audio_codec(
        codec: Union[FfmpegAudioCodec, str]
    ) -> str:
        """
        Sets the general audio codec.

        Code: `-c:a {codec}`
        """
        # We cannot control the big amount of audio codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegAudioCodec.to_enum(codec).value
        except:
            pass

        return f'-c:a {codec}'
    
    @staticmethod
    def video_codec(
        codec: Union[FfmpegVideoCodec, str]
    ) -> str:
        """
        Sets the general video codec.

        Code: `-c:v {codec}`
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-c:v {codec}'

    @staticmethod
    def v_codec(
        codec: Union[FfmpegVideoCodec, str]
    ) -> str:
        """
        Sets the video codec.

        TODO: I don't know exactly the difference between '-c:v {codec}'
        and the '-vcodec' generated in this method. I keep this method
        until I actually find the difference. I don't even know if the
        video codecs I can provide as values are the same as in the other
        method.

        Code: `-vcodec {codec}`
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-vcodec {codec}'

    @staticmethod
    def codec(
        codec: Union[FfmpegVideoCodec, FfmpegAudioCodec, str]
    ) -> str:
        """
        Sets the general codec with '-c {codec}'.

        -c copy indica que se deben copiar los flujos de audio y video sin recodificación, lo que hace que la operación sea rápida y sin pérdida de calidad. TODO: Turn this 'copy' to AudioCodec and VideoCodec (?)

        Code: `-c {codec}`
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter

        # TODO: Validate provided 'codec'
        # TODO: This method has a variation, it can be '-c:a' or '-c:v'
        if not PythonValidator.is_instance_of(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            try:
                codec = FfmpegVideoCodec.to_enum(codec)
            except:
                pass

        if not PythonValidator.is_instance_of(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            try:
                codec = FfmpegAudioCodec.to_enum(codec)
            except:
                pass

        if PythonValidator.is_instance_of(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            codec = codec.value

        return f'-c {codec}'
    
    @staticmethod
    def map(
        map: str
    ) -> str:
        """
        Set input stream mapping.
        -map [-]input_file_id[:stream_specifier][,sync_file_id[:stream_s set input stream mapping

        # TODO: Improve this

        Code: `-map {map}`
        """
        ParameterValidator.validate_mandatory_string('map', map, do_accept_empty = False)

        return f'-map {map}'
    
    @staticmethod
    def filter(
        filter: FfmpegFilter
    ) -> str:
        """
        Sets the expected filter to be used.

        Code: `-filter {filter}`
        """
        filter = FfmpegFilter.to_enum(filter).value

        return f'-filter {filter}'
    
    @staticmethod
    def frame_rate(
        frame_rate: int
    ) -> str:
        """
        Sets the frame rate (Hz value, fraction or abbreviation)

        Code: `-r {frame_rate}`
        """
        ParameterValidator.validate_mandatory_int('frame_rate', frame_rate)
        # TODO: Maybe accept some range (?)

        return f'-r {str(frame_rate)}'
    
    @staticmethod
    def pixel_format(
        format: FfmpegPixelFormat
    ) -> str:
        """
        Set the pixel format.

        Code: `-pix_fmt {format}`
        """
        format = FfmpegPixelFormat.to_enum(format).value

        return f'-pix_fmt {format}'
    
    @staticmethod
    def scale_with_size(
        size: tuple
    ) -> str:
        """
        Set a new size.

        Code: `-vf scale=size[0]:size[1]`
        """
        ParameterValidator.validate_mandatory_tuple('size', size, 2)

        return f'-vf scale={str(int(size[0]))}:{str(int(size[1]))}'

    @staticmethod
    def scale_with_factor(
        w_factor: float,
        h_factor: float
    ) -> str:
        """
        Set a new size multiplying by a factor.

        Code: `-vf "scale=iw*w_factor:ih*h_factor"`
        """
        ParameterValidator.validate_mandatory_float('w_factor', w_factor)
        ParameterValidator.validate_mandatory_float('h_factor', h_factor)

        return f'-vf "scale=iw*{str(w_factor)}:ih*{str(h_factor)}"'

    @staticmethod
    def crop(
        size: tuple,
        origin: tuple
    ) -> str:
        """
        Crop the video to a new with the provided 'size'
        starting with the top left corner at the given
        'origin' position of the original video.
        
        Code: `-vf "crop=size[0]:size[1]:origin[0]:origin[1]"`
        """
        ParameterValidator.validate_mandatory_tuple('size', size, 2)
        ParameterValidator.validate_mandatory_tuple('origin', origin, 2)

        return f"-vf \"crop={str(int(size[0]))}:{str(int(size[1]))}:{str(int(origin[0]))}:{str(int(origin[1]))}\""
    
    @staticmethod
    def seeking(
        seconds: int
    ) -> str:
        """
        Skip the necessary amount of time to go directly
        to the provided 'seconds' time of the input (that
        must be provided after this).

        Code: `-ss 00:00:03`
        """
        ParameterValidator.validate_mandatory_int('seconds', seconds)

        return f'-ss {Date.seconds_to_hh_mm_ss(seconds)}'
    
    @staticmethod
    def to(
        seconds: int
    ) -> str:
        """
        Used with 'seeking' to match the duration we want
        to apply to the new trimmed input. In general, 
        this will be the amount of 'seconds' to be played.

        Code: `-to 00:00:05`
        """
        ParameterValidator.validate_mandatory_int('seconds', seconds)

        return f'-to {Date.seconds_to_hh_mm_ss(seconds)}'


from yta_video_ffmpeg.flag import FfmpegFlag
from typing import Union
from subprocess import run


class FfmpegCommand:
    """
    Class to represent a command to be built and
    executed by the FfmpegHandler.

    A valid example of a command is built like this:
    
    FfmpegCommand([
        FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
        FfmpegFlag.safe_routes(0),
        FfmpegFlag.overwrite,
        FfmpegFlag.frame_rate(frame_rate),
        FfmpegFlag.input(concat_filename),
        FfmpegFlag.video_codec(FfmpegVideoCodec.QTRLE),
        FfmpegFlag.pixel_format(pixel_format), # I used 'argb' in the past
        output_filename
    ])
    """
    args: list[Union[FfmpegFlag, any]] = None
    
    def __init__(
        self,
        args: list[Union[FfmpegFlag, any]]
    ):
        # TODO: Validate args
        self.args = args

    def run(
        self
    ):
        """
        Run the command.
        """
        run(self.__str__())

    def __str__(
        self
    ) -> str:
        """
        Turn the command to a string that can be directly
        executed as a ffmpeg command.
        """
        # TODO: Clean args (?)
        # Remove 'None' args, our logic allows them to make it easier
        args = [
            arg
            for arg in self.args
            if arg is not None
        ]

        return f"ffmpeg {' '.join(args)}"



"""
TODO: Check the 'yta_video_opengl' because 
I'm using pyav there, which is using ffmpeg
in the background, so this is unnecessary.
"""
from dataclasses import dataclass

import numpy as np
import subprocess
import json


@dataclass
class FfmpegVideoInfo:
    """
    Dataclass to hold the information about
    a video read with Ffmpeg.
    """

    def __init__(
        self,
        filename: str,
        json_data: any
    ):
        self.filename: str = filename
        """
        The video file name.
        """

        format_data = json_data.get('format', {})
        self.duration = format_data.get('duration', 0)
        """
        The duration of the video in seconds.
        """
        self.size = format_data.get('size', 0)
        self.bit_rate = format_data.get('bit_rate', 0)
        self.video = {
            'codec': None,
            'size': None,
            'fps': None,
            'pix_fmt': None
        }
        self.audio = {}

        for stream in json_data.get('streams', []):
            if stream.get('codec_type') == 'video':
                # FPS se calcula a partir de r_frame_rate
                fps_str = stream.get("r_frame_rate", "0/0")
                try:
                    num, den = map(int, fps_str.split("/"))
                    fps = num / den if den != 0 else 0
                except ValueError:
                    fps = 0

                self.video_codec: str = stream.get('codec_name', None)
                self.size: tuple[int, int] = (
                    int(stream.get('width', 0)),
                    int(stream.get('height', 0))
                )
                self.fps = fps
                self.pixel_format = stream.get('pix_fmt')

                self.has_alpha = False
                if (
                    self.pixel_format and
                    any(
                        a in self.pixel_format
                        for a in ['a', 'alpha']
                    )
                ):
                    # Ej: yuva420p, rgba, bgra...
                    if self.pixel_format.startswith(('yuva', 'rgba', 'bgra', 'argb', 'ya')):
                        self.has_alpha = True

                # Rotation
                # TODO: None or 0 (?)
                rotation = None
                if 'tags' in stream and 'rotate' in stream['tags']:
                    try:
                        rotation = int(stream['tags']['rotate'])
                    except ValueError:
                        pass

                # Caso 2: side_data_list con "rotation"
                for side_data in stream.get('side_data_list', []):
                    if 'rotation' in side_data:
                        try:
                            rotation = int(side_data['rotation'])
                        except ValueError:
                            pass

                self.rotation = rotation
            elif stream.get('codec_type') == 'audio':
                self.audio_codec: str = stream.get('codec_name', None)
                self.number_of_channels = int(stream.get('channels', 0))
                self.sample_rate = int(stream.get('sample_rate', 0))

class FfmpegReader:
    """
    Class to wrap functionality related to
    reading videos.
    """

    @property
    def process(
        self
    ):
        if not hasattr(self, '_process'):
            self._process = subprocess.Popen(
                [
                    "ffmpeg",
                    "-i", self.filename,
                    "-f", "image2pipe",
                    "-pix_fmt", self.pixel_format,
                    "-vcodec", "rawvideo",
                    "-"
                ],
                stdout = subprocess.PIPE,
                stderr = subprocess.PIPE
            )

        return self._process
    
    @property
    def width(
        self
    ) -> int:
        return self.metadata.size[0]
    
    @property
    def height(
        self
    ) -> int:
        return self.metadata.size[1]

    def __init__(
        self,
        filename: str,
        pixel_format: str = 'rgba'
    ):
        self.filename: str = filename
        """
        The filename of the video we are reading.
        """
        self.metadata: FfmpegVideoInfo = FfmpegReader.get_video_metadata(filename)
        """
        The information about the video.
        """

        # TODO: Why this 'pixel_format' (?)
        self.pixel_format = pixel_format
        self.process = None

    def read_frame(
        self
    ):
        frame_size = self.width * self.height * len(self.pixel_format)
        raw_frame = self.process.stdout.read(frame_size)

        if len(raw_frame) != frame_size:
            # End of the video
            return None

        frame = np.frombuffer(raw_frame, np.uint8).reshape((self.height, self.width, len(self.pix_fmt)))
        return frame
    
    def close(self):
        if self.process:
            self.process.stdout.close()
            self.process.terminate()
            self.process.wait()
            self.process = None

    # TODO: Maybe move this to a utils
    @staticmethod
    def get_video_metadata(
        filename: str
    ) -> FfmpegVideoInfo:
        cmd = [
            "ffprobe",
            # Errors only
            "-v", "error",
            # General data (duration, bitrate, etc.)
            "-show_format",
            # Stream info (video, audio, etc.)
            "-show_streams",
            "-print_format", "json",# Salida en JSON
            filename
        ]
        
        result = subprocess.run(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, text = True)
        
        if result.returncode != 0:
            raise RuntimeError(f"Error ejecutando ffprobe: {result.stderr}")
        
        json_data = json.loads(result.stdout)

        return FfmpegVideoInfo(filename, json_data)



"""
This is a very nice tool to interact with
the powerfull ffmpeg but making it easier.

Check these links below:
- https://www.reddit.com/r/ffmpeg/comments/ks8zfs/comment/gieu7x6/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
- https://stackoverflow.com/questions/38368105/ffmpeg-custom-sequence-input-images/51618079#51618079
- https://stackoverflow.com/a/66014158
"""
from yta_video_ffmpeg.flag import FfmpegFlag
from yta_video_ffmpeg.command import FfmpegCommand
from yta_image_utils.resize import get_cropping_points_to_keep_aspect_ratio
from yta_video_utils.dimensions import get_video_size
from yta_positioning.coordinate import NORMALIZATION_MAX_VALUE, NORMALIZATION_MIN_VALUE
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from yta_file.handler import FileHandler
from yta_constants.file import FileType, FileExtension
from yta_constants.video import FfmpegAudioCodec, FfmpegFilter, FfmpegPixelFormat, FfmpegVideoCodec, FfmpegVideoFormat
from yta_programming.output import Output
from yta_temp import Temp
from typing import Union
from subprocess import run


class _VideoFfmpegHandlerEncoding:
    """
    Class to wrap the functionality related
    to encoding videos and audios. Useful
    when we need to transform a video into
    a I-frame video to be able to edit it
    more efficiently.
    """

    def to_prores(
        self,
        video_filename: str,
        output_filename: str
    ):
        """
        Encode the 'video_filename' to a ProRes
        video that is perfect for edition (it has
        I-frames).

        The recommended extensions are:
        - `.mov`
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileExtension.MOV)

        """
        Help:
        -c:v prores_ks → ProRes coder
        -profile:v 3 → 'standard' profile (HQ is 4, bigger size)
        -c:a copy → copy without recompressing
        """

        # Valid example:
        # ffmpeg -i input.mp4 -c:v prores_ks -profile:v 3 -c:a copy output_prores.mov
        run(f'ffmpeg -i {video_filename} -c:v prores_ks -profile:v 3 -c:a copy {output_filename}')

        return output_filename
    
    def to_dnxhr(
        self,
        video_filename: str,
        output_filename: str
    ):
        """
        Encode the 'video_filename' to a DNxHR
        video that is perfect for edition (it has
        I-frames).

        The recommended extensions are:
        - `.mov`
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileExtension.MOV)

        # Valid example:
        # ffmpeg -i input.mp4 -c:v dnxhd -b:v 36M -c:a copy output_dnxhr.mov
        run(f'ffmpeg -i {video_filename} -c:v dnxhd -b:v 36M -c:a copy {output_filename}')

        return output_filename
    
    def to_mjpeg(
        self,
        video_filename: str,
        output_filename: str
    ):
        """
        Encode the 'video_filename' to a MJPEG
        video that is perfect for edition (it has
        I-frames).

        The recommended extensions are:
        - `.avi`
        - `.mov`
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileExtension.AVI)

        """
        Help:
        -c:v mjpeg → Motion JPEG (MJPEG) codec
        -q:v 3 → JPEG quality (1 = best, 31 = worst)
        -c:a copy → copy without recompressing
        """

        # Valid example:
        # ffmpeg -i input.mp4 -c:v mjpeg -q:v 3 -c:a copy output_mjpeg.avi
        run(f'ffmpeg -i {video_filename} -c:v mjpeg -q:v 3 -c:a copy {output_filename}')

        return output_filename

class _VideoFfmpegHandlerInterpolation:
    """
    Class to wrap the functionality related
    to interpolation.
    """

    def interpolate_slow(
        self,
        video_filename: str,
        fps: int,
        output_filename: str
    ):
        """
        Interpolate the 'video_filename' video to
        have the given 'fps'.

        This process can take a lot of time but 
        the result should be very nice. A test with
        a 20s video has taken 21 minutes in my
        sh*tty laptop.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        # Valid example:
        # 'ffmpeg -i input.mp4 -vf "minterpolate=fps=60" output.mp4'
        run(f'ffmpeg -i {video_filename} -vf "minterpolate=fps={str(fps)}" {output_filename}')

        return output_filename

    def interpolate_normal(
        self,
        video_filename: str,
        fps: int,
        output_filename: str
    ):
        """
        Generate intermediate frames at the provided
        'fps' with simplified interpolation.

        Soft results, a bit slower than level 1 but
        better results.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        # Valid example:
        # 'ffmpeg -i input.mp4 -vf "minterpolate=fps=60" -c:v libx264 -preset fast -crf 18 output.mp4'
        run(f'ffmpeg -i {video_filename} -vf "minterpolate=fps={str(fps)}" -c:v libx264 -preset ultrafast -crf 23 {output_filename}')

        return output_filename
    
    def interpolate_fast(
        self,
        video_filename: str,
        fps: int,
        output_filename: str
    ):
        """
        Make a blend between frames to be smooth,
        but scenes with a lot of movement can be
        blurry.

        Depends a lot on the input, but could be a
        good and not too slow option.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        # Valid example:
        # 'ffmpeg -i input.mp4 -vf "fps=60,tblend=all_mode=average" output.mp4'
        run(f'ffmpeg -i {video_filename} -vf "fps={str(fps)},tblend=all_mode=average" {output_filename}')

        return output_filename
        
    def interpolate_instant(
        self,
        video_filename: str,
        fps: int,
        output_filename: str
    ):
        """
        Modify the amount of fps to the given 'fps'
        parameter but with no improvement.

        Not a good result but the fastest one.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        # Valid example:
        # TODO: This -crf 18 below is apparently max
        # quality, so it is very slow
        # 'ffmpeg -i input.mp4 -vf "fps=60" -c:v libx264 -preset fast -crf 18 output.mp4'
        # run(f'ffmpeg -i {video_filename} -vf "fps={str(fps)}" -c:v libx264 -preset fast -crf 18 {output_filename}')
        # Use this for a fast result:
        # ffmpeg -i input.mp4 -vf "fps=60" -c:v libx264 -preset ultrafast -crf 23 output.mp4
        run(f'ffmpeg -i {video_filename} -vf "fps={str(fps)}" -c:v libx264 -preset ultrafast -crf 23 {output_filename}')

        return output_filename
        

class _VideoFfmpegHandler:
    """
    Class to wrap the functionality related
    to videos.
    """

    def __init__(
        self
    ):
        self.interpolation: _VideoFfmpegHandlerInterpolation = _VideoFfmpegHandlerInterpolation()
        """
        Shortcut to the functionality related to
        video interpolation.
        """
        self.encoding: _VideoFfmpegHandlerEncoding = _VideoFfmpegHandlerEncoding()
        """
        Shortcut to the functionality related to
        video encoding.
        """
        
    # TODO: Check this one below
    def extract_audio_deprecated(
        self,
        video_filename: str,
        codec: FfmpegAudioCodec = None,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        # TODO: This has not been tested yet.

        Pro tip: You can read the return with AudioParser.as_audioclip
        method.

        This method returns the filename of the file that has been
        generated as a the audio of the provided video.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.AUDIO)

        codec = (
            FfmpegAudioCodec.to_enum(codec)
            if codec is not None else
            None
        )

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            (
                FfmpegFlag.audio_codec(codec)
                if codec else
                None
            ),
            output_filename
        ]).run()

        return output_filename
    
    def extract_audio(
        self,
        video_filename: str,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        Pro tip: You can read the return with AudioParser.as_audioclip
        method.

        This method returns the filename of the file that has been
        generated as a the audio of the provided video.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.AUDIO)
        # TODO: Verify valid output_filename extension

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.map('0:1'),
            output_filename
        ]).run()

        return output_filename
    
    def get_best_thumbnail(
        self,
        video_filename: str,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Gets the best thumbnail of the provided 'video_filename'.

        Pro tip: You can read the return with ImageParser.to_pillow
        method.

        This method returns the filename of the file that has been
        generated as a the thumbnail of the provided video.
        """
        validate_video_filename(video_filename)
        output_filename = Output.get_filename(output_filename, FileType.IMAGE)

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.filter(FfmpegFilter.THUMBNAIL),
            output_filename
        ]).run()

        return output_filename
    
    def concatenate(
        self,
        video_filenames: list[str],
        output_filename: str = None
    ) -> str:
        """
        Concatenates the provided 'video_filenames' in the order in
        which they are provided.

        Using Ffmpeg is very useful when trying to concatenate similar
        videos (the ones that we create always with the same 
        specifications) because the codecs are the same so the speed
        is completely awesome.

        Pro tip: You can read the return with VideoParser.to_moviepy
        method.

        This method returns the filename of the file that has been
        generated as a concatenation of the provided ones.
        """
        for video_filename in video_filenames:
            validate_video_filename(video_filename)

        output_filename = Output.get_filename(output_filename, FileType.VIDEO)
        concat_filename = FfmpegHandler.write_concat_file(video_filenames)

        FfmpegCommand([
            FfmpegFlag.overwrite,
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.codec('copy'),
            output_filename
        ]).run()

        return output_filename
    
    def resize(
        self,
        video_filename: str,
        size: tuple,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Resize the provided 'video_filename', by keeping
        the aspect ratio (cropping if necessary), to the
        given 'size' and stores it locally as
        'output_filename'.

        This method returns the generated file filename.

        See more: 
        https://www.gumlet.com/learn/ffmpeg-resize-video/
        """
        validate_video_filename(video_filename)
        ParameterValidator.validate_mandatory_tuple('size', size, None)
        # TODO: I think we don't need this with the 'Output.get_filename'
        ParameterValidator.validate_mandatory_string('output_filename', output_filename, do_accept_empty = False)
        
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        # Validate that 'size' is a valid size
        # TODO: This code is a bit strange, but was refactored from the
        # original one that was in 'yta_multimedia' to remove the
        # dependency. Maybe update it?
        if not PythonValidator.is_numeric_tuple_or_list_or_array_of_2_elements_between_values(size, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE):
            # TODO: Raise error
            raise Exception(f'The provided size parameter is not a tuple or array, or does not have 2 elements that are numbers between {str(NORMALIZATION_MIN_VALUE)} and {str(NORMALIZATION_MAX_VALUE)}.')

        w, h = get_video_size(video_filename)

        if (w, h) == size:
            # No need to resize, we just copy it to output
            FileHandler.copy_file(video_filename, output_filename)
        else:
            # First, we need to know if we need to scale it
            original_ratio = w / h
            new_ratio = size[0] / size[1]

            new_size = (
                (w * (size[1] / h), size[1])
                # Original video is wider than the expected one
                if original_ratio > new_ratio else
                # Original video is higher than the expected one
                (size[0], h * (size[0] / w))
                if original_ratio < new_ratio else
                (size[0], size[1])
            )

            tmp_filename = Temp.get_wip_filename('tmp_ffmpeg_scaling.mp4')

            # Scale to new dimensions
            FfmpegCommand([
                FfmpegFlag.input(video_filename),
                FfmpegFlag.scale_with_size(new_size),
                tmp_filename
            ]).run()

            # Now, with the new video resized, we look for the
            # cropping points we need to apply and we crop it
            top_left, _ = get_cropping_points_to_keep_aspect_ratio(new_size, size)

            # Second, we need to know if we need to crop it
            FfmpegCommand([
                FfmpegFlag.input(tmp_filename),
                FfmpegFlag.crop(size, top_left),
                FfmpegFlag.overwrite,
                output_filename
            ]).run()

        return output_filename
    
    def trim(
        self,
        video_filename: str,
        start_seconds: int,
        duration_seconds: int,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Trims the provided 'video_filename' and generates a
        shorter resource that is stored as 'output_filename'.
        This new resource will start from 'start_seconds' and
        last the provided 'duration_seconds'.

        This method returns the generated file filename.

        Thank you:
        https://www.plainlyvideos.com/blog/ffmpeg-trim-videos
        https://trac.ffmpeg.org/wiki/Seeking
        """
        validate_video_filename(video_filename)
        ParameterValidator.validate_mandatory_positive_number('start_seconds', start_seconds, do_include_zero = True)
        ParameterValidator.validate_mandatory_positive_number('duration_seconds', duration_seconds, do_include_zero = True)
        
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)

        command = FfmpegCommand([
            FfmpegFlag.seeking(start_seconds),
            FfmpegFlag.input(video_filename),
            FfmpegFlag.to(duration_seconds),
            FfmpegFlag.codec(FfmpegVideoCodec.COPY),
            FfmpegFlag.overwrite,
            output_filename
        ])
        # TODO: Remove this command when confirmed
        print(command)
        command.run()

        #ffmpeg_command = f'-ss 00:02:05 -i {video} -to 00:03:10 -c copy video-cutted-ffmpeg.mp4'
        return output_filename
    
    def set_audio(
        self,
        video_filename: str,
        audio_filename: str,
        output_filename: Union[str, None] = None
    ):
        """
        TODO: This method has not been properly tested yet.

        Set the audio given in the 'audio_filename' in the also
        provided video (in 'video_filename') and creates a new
        file containing the video with the audio.

        This method returns the generated file filename.
        """
        validate_video_filename(video_filename)
        validate_audio_filename(audio_filename)
        # TODO: I think we don't need this with the 'Output.get_filename'
        ParameterValidator.validate_mandatory_string('output_filename', output_filename)
        
        output_filename = Output.get_filename(output_filename, FileType.VIDEO)
        
        # cls.run_command([
        #     FfmpegFlag.input(video_filename),
        #     FfmpegFlag.input(audio_filename),
        #     output_filename
        # # TODO: Unfinished
        # ])

        # TODO: Is this actually working (?)
        run(f"ffmpeg -i {video_filename} -i {audio_filename} -c:v copy -c:a aac -strict experimental -y {output_filename}")

        return output_filename
        
        # Apparently this is the equivalent command according
        # to ChatGPT, but maybe it doesn't work
        # ffmpeg -i input_video -i input_audio -c:v copy -c:a aac -strict experimental -y output_filename

        # There is also a post that says this:
        # ffmpeg -i input.mp4 -i input.mp3 -c copy -map 0:v:0 -map 1:a:0 output.mp4
        # in (https://superuser.com/a/590210)


        # # TODO: What about longer audio than video (?)
        # # TODO: This is what was being used before FFmpegHandler
        # input_video = ffmpeg.input(video_filename)
        # input_audio = ffmpeg.input(audio_filename)

        # ffmpeg.concat(input_video, input_audio, v = 1, a = 1).output(output_filename).run(overwrite_output = True)
        
class _AudioFfmpegHandler:
    """
    Class to wrap the functionality related
    to audios.
    """

    def __init__(
        self
    ):
        pass

    # TODO: Implement something for audio files
        
class _ImageFfmpegHandler:
    """
    Class to wrap the functionality related
    to images.
    """

    def __init__(
        self
    ):
        pass

    def concatenate(
        self,
        image_filenames: list[str],
        frame_rate = 60,
        pixel_format: FfmpegPixelFormat = FfmpegPixelFormat.YUV420p,
        output_filename: Union[str, None] = None
    ) -> str:
        """
        Concatenates the provided 'image_filenames' in the order in
        which they are provided.

        Using Ffmpeg is very useful when trying to concatenate similar
        images because the speed is completely awesome.

        Pro tip: You can read the return with VideoParser.to_moviepy().

        This method returns the filename of the file that has been
        generated as a concatenation of the provided ones.
        """
        for image_filename in image_filenames:
            validate_image_filename(image_filename)

        output_filename = Output.get_filename(output_filename, FileType.VIDEO)
        concat_filename = FfmpegHandler.write_concat_file(image_filenames)

        # TODO: Should we check the pixel format or give freedom (?)
        # pixel_format = FfmpegPixelFormat.to_enum(pixel_format)

        FfmpegCommand([
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.overwrite,
            FfmpegFlag.frame_rate(frame_rate),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.video_codec(FfmpegVideoCodec.QTRLE),
            FfmpegFlag.pixel_format(pixel_format), # I used 'argb' in the past
            output_filename
        ]).run()

        return output_filename

class FfmpegHandler:
    """
    Class to simplify and encapsulate ffmpeg functionality.
    """

    def __init__(
        self
    ):
        self.video: _VideoFfmpegHandler = _VideoFfmpegHandler()
        """
        Shortcut to the functionality related to
        videos.
        """
        self.audio: _AudioFfmpegHandler = _AudioFfmpegHandler()
        """
        Shortcut to the functionality related to
        audios.
        """
        self.image: _ImageFfmpegHandler = _ImageFfmpegHandler()
        """
        Shortcut to the functionality related to
        images.
        """

    def write_concat_file(
        self,
        filenames: str
    ) -> str:
        """
        Write the files to concat in a temporary text file with
        the required format and return that file filename. This
        is required to use different files as input.

        This method returns the created file filename that 
        includes the list with the 'filenames' provided ready
        to be concatenated.
        """
        text = '\n'.join(
            f"file '{filename}'"
            for filename in filenames
        )

        # TODO: Maybe this below is interesting for the 'yta_general_utils.file.writer'
        # open('concat.txt', 'w').writelines([('file %s\n' % input_path) for input_path in input_paths])
        filename = Temp.get_wip_filename('concat_ffmpeg.txt')
        FileHandler.write_str(filename, text)

        return filename

    def run_command(
        self,
        command: Union[list[FfmpegFlag, any], FfmpegCommand]
    ) -> None:
        """
        Runs the provided ffmpeg 'command'.
        """
        command = (
            FfmpegCommand(command)
            if not PythonValidator.is_instance_of(command, FfmpegCommand) else
            command
        )

        command.run()

def validate_video_filename(
    video_filename: str
) -> None:
    """
    Validate if the provided 'video_filename'
    parameter is a string and a valid video file
    (based on its extension).
    """
    ParameterValidator.validate_mandatory_string('video_filename', video_filename, do_accept_empty = False)

    # TODO: If possible (and no dependency issue) check 
    # the content to validate it is parseable as video
    if not FileHandler.is_video_file(video_filename):
        raise Exception('The provided "video_filename" is not a valid video file name.')
    
def validate_audio_filename(
    audio_filename: str
) -> None:
    """
    Validate if the provided 'audio_filename'
    parameter is a string and a valid audio file
    (based on its extension).
    """
    ParameterValidator.validate_mandatory_string('audio_filename', audio_filename, do_accept_empty = False)

    # TODO: If possible (and no dependency issue) check 
    # the content to validate it is parseable as audio
    if not FileHandler.is_audio_file(audio_filename):
        raise Exception('The provided "audio_filename" is not a valid audio file name.')
    
def validate_image_filename(
    image_filename: str
):
    """
    Validate if the provided 'image_filename'
    parameter is a string and a valid image file
    (based on its extension).
    """
    ParameterValidator.validate_mandatory_string('image_filename', image_filename, do_accept_empty = False)

    # TODO: If possible (and no dependency issue) check 
    # the content to validate it is parseable as image
    if not FileHandler.is_image_file(image_filename):
        raise Exception('The provided "image_filename" is not a valid image file name.')

    # TODO: Keep going